"""
     ____  ______
    / __ \/_  __/___ _____ ______
   / / / / / / / __ `/ __ `/ ___/
  / /_/ / / / / /_/ / /_/ (__  )
 /_____/ /_/  \__,_/\__, /____/
                   /____/

Directory tags for lazy programmers.
By Joohwan Oh (github.com/joowani/dtags)
"""

import os

LOGO = __doc__

# Determine the home directory on runtime
HOME = os.path.expanduser('~')

# Configuration directory (i.e. ~/.dtags)
CFG_DIR = os.path.join(HOME, '.dtags')

# File containing just the tag names for tab completion
TAGS_FILE = os.path.join(CFG_DIR, 'tags')

# File containing the path-to-tag mapping
MAPPING_FILE = os.path.join(CFG_DIR, 'mapping')



from __future__ import unicode_literals, print_function

import os
import io
import sys
import shlex
import atexit
import signal
import subprocess as sp
from collections import defaultdict
from tempfile import NamedTemporaryFile as TempFile

from dtags import style, CFG_DIR, MAPPING_FILE, TAGS_FILE
from dtags.commands import d
from dtags.chars import get_invalid_tag_chars
from dtags.config import load_mapping, save_mapping, parse_mapping
from dtags.shells import SUPPORTED_SHELLS
from dtags.utils import close_stdio, abort, finish, expand, rm_files
from dtags.version import VERSION

USAGE = """Usage:
  dtags list [<dir>|<tag>...]
  dtags reverse [<dir>|<tag>...]
  dtags edit
  dtags clean
  dtags commands
  dtags --help
  dtags --version
"""
DESCRIPTION = """
Commands:
  list       Display directories-to-tags mapping
  reverse    Display tags-to-directories mapping
  edit       Edit tags and directories via editor
  clean      Remove invalid tags and directories
  commands   List all available dtags commands

Arguments:
  dir        The directory path
  tag        The directory tag

Options:
  --help     Display the help menu
  --version  Display the version"""

EDIT_HELP_COMMENTS = """# Format: <dir>,<tag>,<tag>,<tag>...
# <dir>: the directory path
# <tag>: the directory tag
#
# Examples:
# /home/user/app/backend/mysql,app,backend,mysql
# /home/user/app/backend/redis,app,backend,redis
# /home/user/app/frontend/web,app,frontend,web
"""

COMMANDS = """t: tag a specified directory
u: untag a specified directory
d: go to the specified directory
e: execute command in directories
p: execute command in parallel
"""


def _edit(args):
    """Edit the mapping directly using an editor."""
    if args:
        abort(USAGE + 'dtags: too many arguments')
    try:
        with TempFile(
            mode='w+t', delete=False, prefix='mapping.', dir=CFG_DIR
        ) as tfile:
            with io.open(MAPPING_FILE, 'rt') as mapping_file:
                tfile.write(EDIT_HELP_COMMENTS + mapping_file.read())
            tfile.flush()
    except (OSError, IOError) as err:
        abort('dtags: failed to edit mapping: {}'.format(err), err.errno)
    else:
        editor = shlex.split(os.environ.get('EDITOR'))
        if not editor:
            abort('dtags: undefined environment variable: ' +
                  style.bad('EDITOR'))
        try:
            sp.check_call(editor + [tfile.name])
        except sp.CalledProcessError as err:
            abort('dtags: failed to edit mapping: {}'.format(err.message))
        else:
            mapping, excluded = parse_mapping(tfile.name)
            save_mapping(mapping)
            rm_files(tfile.name)
            if excluded:
                print('Cleaned the following entries:\n' + excluded + '\n')
            finish('New entries saved successfully')


def _clean(args):
    """Clean stale and/or invalid directories."""
    if args:
        abort(USAGE + 'dtags: too many arguments')
    mapping, excluded = load_mapping()
    if excluded:
        save_mapping(mapping)
        finish('Cleaned the following entries:\n' + excluded)
    else:
        save_mapping(mapping)
        finish('Nothing to clean')


def _list(targets=None, reverse=False):
    """List directories and tags.

    :param targets: the list of directories or tags to highlight
    :param reverse: whether to display the reverse mapping or not
    """
    if targets is None:
        targets = set()
    else:
        targets = set(targets)
        additional_targets = set()
        for target in targets:
            if target.startswith('-'):
                abort(USAGE + 'dtags: invalid argument: ' + style.bad(target))
            else:
                additional_targets.add(expand(target))
        targets.update(additional_targets)
    mapping, excluded = load_mapping()
    if excluded:
        print(
            'Found invalid entries in the mapping:\n' + excluded +
            '\nRun ' + style.cmd('dtags clean') + ' to remove them' + '\n'
        )
    if not mapping:
        finish('Nothing to list')
    msgs = []
    if reverse:
        if targets:
            mapping = {
                tag: paths for tag, paths in mapping.items()
                if tag in targets or paths & targets
            }
        for tag in sorted(mapping):
            lines = [style.tag(tag, tag in targets)]
            for path in sorted(mapping[tag]):
                lines.append(style.path(path, path in targets))
            msgs.append('\n'.join(lines))
        finish('\n\n'.join(msgs) if msgs else 'Nothing to list')
    else:
        rmapping = defaultdict(set)
        for tag, paths in mapping.items():
            for path in paths:
                rmapping[path].add(tag)
        if targets:
            rmapping = {
                path: tags for path, tags in rmapping.items()
                if path in targets or tags & targets
            }
        for path in sorted(rmapping):
            tags = ' '.join(
                style.tag(tag, tag in targets)
                for tag in sorted(rmapping[path])
            )
            msgs.append(style.path(path, path in targets) + ' ' + tags)
        finish('\n'.join(msgs) if msgs else 'Nothing to list')


def _commands(args):
    """Display all available commands."""
    if args:
        abort(USAGE + 'dtags: too many arguments')
    print(COMMANDS)


def manage():
    """Manage directory tags."""
    atexit.register(close_stdio)
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    args = sys.argv[1:]
    if not args:
        _list()
    head, tail = args[0], args[1:]
    if head == '--help':
        finish(USAGE + DESCRIPTION)
    elif head == '--version':
        finish('Version ' + VERSION)
    elif head == 'edit':
        _edit(tail)
    elif head == 'clean':
        _clean(tail)
    elif head == 'list':
        _list(tail, reverse=False)
    elif head == 'reverse':
        _list(tail, reverse=True)
    elif head == 'commands':
        _commands(tail)
    else:
        abort(USAGE + 'dtags: invalid argument: ' + style.bad(head))


def activate():
    """Write the shell runtime configuration to stdout."""
    atexit.register(close_stdio)
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    args = sys.argv[1:]
    if len(args) >= 2:
        abort(USAGE + 'dtags: too many arguments')
    if len(args) == 1:
        shell_name = args[0]
        config = SUPPORTED_SHELLS.get(shell_name)
        if config is None:
            abort('dtags: unsupported shell: ' + style.bad(shell_name))
    else:
        shell_path = os.environ.get('SHELL')
        if shell_path is None:
            abort('dtags: undefined environment variable: ' +
                  style.bad('SHELL'))
        shell_name = None
        for supported_shell_name in SUPPORTED_SHELLS:
            if supported_shell_name in shell_path:
                shell_name = supported_shell_name
        config = SUPPORTED_SHELLS.get(shell_name)
        if config is None:
            abort('dtags: unsupported shell: ' + style.bad(shell_path))
    finish(config.format(
        mapping_file=MAPPING_FILE,
        tags_file=TAGS_FILE,
        version=VERSION,
        usage=d.USAGE,
        description=d.DESCRIPTION,
        arg_err_tty=d.ARG_ERR_TTY,
        dest_err_tty=d.DEST_ERR_TTY,
        input_err_tty=d.INPUT_ERR_TTY,
        index_err_tty=d.INDEX_ERR_TTY,
        prompt_tty=d.PROMPT_TTY,
        choice_tty=d.CHOICE_TTY,
        arg_err=d.ARG_ERR,
        dest_err=d.DEST_ERR,
        input_err=d.INPUT_ERR,
        index_err=d.INDEX_ERR,
        prompt=d.PROMPT,
        choice=d.CHOICE,
    ))


def refresh():
    """Write the shell declare variable statements to stdout."""
    atexit.register(close_stdio)
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    args = sys.argv[1:]
    if len(args) >= 2:
        abort(USAGE + 'dtags: too many arguments')
    if len(args) == 1:
        shell_name = args[0]
        config = SUPPORTED_SHELLS.get(shell_name)
        if config is None:
            abort('dtags: unsupported shell: ' + style.bad(shell_name))
    else:
        shell_path = os.environ.get('SHELL')
        if shell_path is None:
            abort('dtags: undefined environment variable: ' +
                  style.bad('SHELL'))
        shell_name = None
        for supported_shell_name in SUPPORTED_SHELLS:
            if supported_shell_name in shell_path:
                shell_name = supported_shell_name
        config = SUPPORTED_SHELLS.get(shell_name)
        if config is None:
            abort('dtags: unsupported shell: ' + style.bad(shell_path))

    assign_string = 'set -g {} "{}"' if shell_name == 'fish' else '{}="{}"'
    mapping, _ = load_mapping()
    print('\n'.join(
        assign_string.format(tag.replace('-', '_'), paths.pop())
        for tag, paths in mapping.items()
        if len(paths) == 1 and
        tag.replace('-', '_') not in os.environ and
        not get_invalid_tag_chars(tag)
    ))







from __future__ import unicode_literals, print_function

import os
import sys
import atexit
import signal
import tempfile
import collections
import subprocess as sp

from dtags import style
from dtags.config import load_mapping
from dtags.utils import close_stdio, abort, finish, expand
from dtags.version import VERSION

temp_file = None
process = None
processes = []

USAGE = """Usage:
  p [-i] <targets> <command> [<arg>...]
  p --help
  p --version
"""
DESCRIPTION = """
Arguments:
  targets     Comma separated directory tags or paths
  command     The command to execute
  arg         The command arguments

Options:
  -i          Execute the commands via interactive shell
  --help      Display the help menu
  --version   Display the version

Execute commands in one or more directories in parallel.
Multiple targets can be given by separating them with
commas and without whitespaces."""


def _send_sigterm(child_process):
    """Send SIGTERM to the specified child process."""
    try:
        os.killpg(os.getpgid(child_process.pid), signal.SIGKILL)
    except OSError as kill_error:
        # If the process is already dead, nothing needs to be done
        if kill_error.errno != 3:  # ESRCH
            raise kill_error


def _cleanup_resources():
    """Kill all child processes and close any open files."""
    global temp_file, process, processes

    if temp_file is not None:
        temp_file.close()
    if process is not None:
        _send_sigterm(process)
    for _, _, child_process, open_file in processes:
        _send_sigterm(child_process)
        if open_file is not None:
            open_file.close()
    close_stdio()


def _handle_signal(signum, *_):
    """Send SIGTERM to all child processes."""
    _cleanup_resources()
    sys.exit(127 + signum)


def main():
    global temp_file, process, processes

    atexit.register(_cleanup_resources)
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    mapping, excluded = load_mapping()
    if excluded:
        print(
            'Found invalid entries in the mapping:\n' +
            excluded + '\nRun ' + style.cmd('dtags clean') +
            ' to remove them' + '\n'
        )
    # Parse the optional arguments
    interactive = False
    args = sys.argv[1:]
    if not args:
        finish(USAGE + DESCRIPTION)
    head, tail = args[0], args[1:]
    if head == '--help':
        finish(USAGE + DESCRIPTION)
    elif head == '--version':
        finish('Version ' + VERSION)
    elif head == '-i' and not tail:
        abort(USAGE + 'p: missing argument: ' + style.bad('<targets>'))
    elif head == '-i' and tail:
        interactive = True
        head, tail = tail[0], tail[1:]
    elif head.startswith('-'):
        abort(USAGE + 'p: invalid argument: ' + style.bad(head))

    # Parse the positional arguments
    if not tail:
        abort(USAGE + 'p: missing argument: ' + style.bad('<command>'))
    directories = collections.defaultdict(set)
    for target in head.split(','):
        if not target:
            continue
        elif target in mapping:
            for directory in sorted(mapping[target]):
                directories[directory].add(target)
        else:
            path = expand(target)
            if os.path.isdir(path):
                directories[path].add(None)
            else:
                abort(USAGE + 'p: invalid target: ' + style.bad(target))

    # Check which shell is in use (e.g. zsh, bash, fish)
    shell = os.environ.get('SHELL')
    if shell is None:
        abort('p: undefined environment variable: ' + style.bad('SHELL'))

    if interactive:
        cmd = [shell, '-i', '-c', sp.list2cmdline(tail)]
    else:
        cmd = [shell, '-c', sp.list2cmdline(tail)]

    # Add signal handlers to terminate child processes gracefully
    signal.signal(signal.SIGINT, _handle_signal)
    signal.signal(signal.SIGABRT, _handle_signal)
    signal.signal(signal.SIGTERM, _handle_signal)

    # Execute in parallel and pipe the output to temporary files
    for directory in sorted(directories):
        tags = directories[directory]
        temp_file = tempfile.TemporaryFile(mode='w+t')
        process = sp.Popen(
            cmd,
            cwd=directory,
            stdout=temp_file,
            stderr=temp_file,
            preexec_fn=os.setsid
        )
        processes.append((directory, tags, process, temp_file))

    # Read from the temporary files back to stdout line by line
    sys.stdout.write('\n')
    for directory, tags, process, temp_file in processes:
        status = process.wait()
        tags = [style.tag(tag) for tag in tags if tag]
        sys.stdout.write('{} {}{}\n'.format(
            style.msg('in'), style.path(directory),
            ((' ' + ' '.join(tags)) if tags else '') + style.msg(':')
        ))
        temp_file.seek(0)
        for line in temp_file:
            sys.stdout.write(line)
        temp_file.close()
        if status != 0:
            sys.stdout.write(style.bad('Exit status: {}\n\n'.format(status)))
        else:
            sys.stdout.write('\n\n')



CONFIGURATION = """
unalias dtags > /dev/null 2>&1
unalias t > /dev/null 2>&1
unalias u > /dev/null 2>&1
unalias e > /dev/null 2>&1
unalias p > /dev/null 2>&1
unalias d > /dev/null 2>&1

function dtags() {{
    dtags-manage $@
    if [[ $1 = edit ]] || [[ $1 = clean ]]
    then
        eval "`dtags-refresh zsh`"
    fi
}}

function t() {{
    dtags-t $@ && eval "`dtags-refresh zsh`"
}}

function u() {{
    dtags-u $@ && eval "`dtags-refresh zsh`"
}}

function d() {{
    declare _dtags_usage='{usage}'
    declare _dtags_version='{version}'
    declare _dtags_description='{description}'
    if [[ -t 1 ]]
    then
        declare _dtags_arg_err="{arg_err_tty}"
        declare _dtags_dest_err="{dest_err_tty}"
        declare _dtags_input_err="{input_err_tty}"
        declare _dtags_idx_err="{index_err_tty}"
        declare _dtags_prompt="{prompt_tty}"
        declare _dtags_choice="{choice_tty}"
    else
        declare _dtags_arg_err="{arg_err}"
        declare _dtags_dest_err="{dest_err}"
        declare _dtags_input_err="{input_err}"
        declare _dtags_idx_err="{index_err}"
        declare _dtags_prompt="{prompt}"
        declare _dtags_choice="{choice}"
    fi
    if [[ $# -lt 1 ]]
    then
        cd $HOME
        return 0
    elif [[ $1 = --help ]]
    then
        printf "$_dtags_usage$_dtags_description"
        return 0
    elif [[ $1 = --version ]]
    then
        printf "Version $_dtags_version\n"
        return 0
    elif [[ $1 = - ]]
    then
        cd "$OLDPWD"
        return 0
    elif [[ $1 = -* ]]
    then
        printf "$_dtags_arg_err" "$_dtags_usage" "$1"
        return 2
    elif [[ $# -gt 1 ]]
    then
        printf "%sd: too many arguments\n" "$_dtags_usage"
        return 2
    fi
    declare -a _dtags_dirs
    PREV_IFS=$IFS
    IFS=$'\n'
    _dtags_dirs=($(grep -F ,$1, {mapping_file} | cut -d',' -f1))
    IFS=$PREV_IFS
    declare _dtags_count=${{#_dtags_dirs[@]}}

    if [[ $_dtags_count -eq 0 ]]
    then
        if [[ -d $1 ]]
        then
            declare _dtags_dir="$1"
        else
            printf "$_dtags_dest_err" "$1"
            return 2
        fi
    elif [[ $_dtags_count = 1 ]]
    then
        declare _dtags_dir=${{_dtags_dirs[1]}}
    else
        for i in `seq $_dtags_count`
        do
            printf "$_dtags_choice" "$i" "${{_dtags_dirs[$i]}}"
        done
        printf "$_dtags_prompt" "$_dtags_count"
        declare _dtags_idx=1
        read _dtags_idx
        if [[ ! $_dtags_idx =~ ^-?[0-9]+$ ]]
        then
            printf "$_dtags_input_err" "$_dtags_idx"
            return 2
        elif [[ $_dtags_idx -gt $_dtags_count ]] || [[ $_dtags_idx -lt 1 ]]
        then
            printf "$_dtags_idx_err" "$_dtags_idx"
            return 2
        fi
        declare _dtags_dir=${{_dtags_dirs[$_dtags_idx]}}
    fi
    if [[ -d $_dtags_dir ]]
    then
        cd "$_dtags_dir"
        return 0
    else
        printf "$_dtags_dest_err" "$_dtags_dir"
        return 2
    fi
}}

_dtags() {{
    if [[ CURRENT -eq 2 ]]
    then
        compadd list reverse
        compadd -S '' edit clean commands
    elif [[ $words[2] = list || $words[2] = reverse ]]
    then
        [[ -f "{tags_file}" ]] && compadd -S '' `cat {tags_file}`
        _files -/
    fi
}}

_t() {{
    if [[ CURRENT -eq 2 ]]
    then
        _files -/
    elif [[ -f "{tags_file}" ]] && [[ ! $words[2] = -* ]]
    then
        compadd `cat {tags_file}`
    fi
}}


_u() {{
    if [[ CURRENT -eq 2 ]]
    then
        _files -/
    elif [[ -f "{tags_file}" ]] && [[ ! $words[2] = -* ]]
    then
        compadd `cat {tags_file}`
    fi
}}

_e() {{
    if [[ CURRENT -eq 2 ]]
    then
        [[ -f "{tags_file}" ]] && compadd `cat {tags_file}`
        _files -/
    elif [[ CURRENT -gt 3 ]]
    then
        _files
    fi
}}

_d() {{
    if [[ CURRENT -eq 2 ]]
    then
        if [[ -f "{tags_file}" ]]
        then
            compadd -S '' `cat {tags_file}`
        fi
        _files -/
    fi
}}

compdef _dtags dtags
compdef _t t
compdef _u u
compdef _e e
compdef _e p
compdef _d d

eval "`dtags-refresh zsh`"
"""



from __future__ import unicode_literals, print_function

import sys
import atexit
import signal

from dtags import style
from dtags.config import load_mapping, save_mapping
from dtags.utils import close_stdio, finish, expand, abort
from dtags.version import VERSION

USAGE = """Usage:
  u <dir> [<tag>...]
  u --help
  u --version
"""
DESCRIPTION = """
Arguments:
  dir     The directory path
  tag     The directory tag

Untag the target directory. If no tag names are
given, ALL tags mapped to the target directory
are removed."""


def main():
    atexit.register(close_stdio)
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    mapping, excluded = load_mapping()
    args = sys.argv[1:]
    if not args:
        finish(USAGE + DESCRIPTION)
    head, tail = args[0], args[1:]
    if head == '--help':
        finish(USAGE + DESCRIPTION)
    elif head == '--version':
        finish('Version ' + VERSION)
    path = expand(head)
    tags_removed = set()
    for tag in tail if tail else mapping.keys():
        if path in mapping[tag]:
            mapping[tag].remove(path)
            tags_removed.add(tag)
    if not tags_removed:
        finish('Nothing to do')
    save_mapping(mapping)
    if excluded:
        print('Cleaned the following invalid entries:\n' + excluded + '\n')
    finish(style.path(path) + ' ' + ' '.join(
        style.sign('-') + style.tag(tag) for tag in sorted(tags_removed)
    ))



from __future__ import unicode_literals

import sys

TTY = sys.stdout.isatty()

# Normal color
BLACK = '\033[38;5;240m'
BLUE = '\033[38;5;27m'
BROWN = '\033[38;5;88m'
GRAY = '\033[38;5;249m'
GREEN = '\033[38;5;154m'
ORANGE = '\033[38;5;208m'
PURPLE = '\033[38;5;141m'
RED = '\033[38;5;009m'
SKY = '\033[38;5;39m'
TURQUOISE = '\033[38;5;37m'
YELLOW = '\033[38;5;214m'

# Bold
BLACK_BOLD = '\033[1;38;5;240m'
BLUE_BOLD = '\033[1;38;5;27m'
BROWN_BOLD = '\033[1;38;5;88m'
GRAY_BOLD = '\033[1;38;5;249m'
GREEN_BOLD = '\033[1;38;5;154m'
ORANGE_BOLD = '\033[1;38;5;208m'
PURPLE_BOLD = '\033[1;38;5;141m'
RED_BOLD = '\033[1;38;5;009m'
SKY_BOLD = '\033[1;38;5;39m'
TURQUOISE_BOLD = '\033[1;38;5;37m'
YELLOW_BOLD = '\033[1;38;5;214m'

# Clear format
CLEAR = '\033[0m'


def tag(value, highlight=False, tty=TTY):
    """Style the directory tag name.

    :param value: the directory tag name to style
    :param highlight: True iff the tag is to be highlighted
    :param tty: True iff the output is going to a terminal
    :return: the styled directory tag name
    """
    if tty:
        if highlight:
            return BLACK + '#' + GREEN_BOLD + value + CLEAR
        else:
            return BLACK + '#' + GRAY + value + CLEAR
    return '#' + value


def path(value, highlight=False, tty=TTY):
    """Style the directory path.

    :param value: the directory path to style
    :param highlight: True iff the path is to be highlighted
    :param tty: True iff the output is going to a terminal
    :return: the styled directory path
    """
    if tty:
        if highlight:
            return GREEN_BOLD + value + CLEAR
        else:
            return SKY + value + CLEAR
    return value


def bad_chars(value, tty=TTY):
    """Style the set of invalid characters in directory tags/paths.

    :param value: the set of invalid characters to style
    :param tty: True iff the output is going to a terminal
    :return: the styled invalid characters
    """
    if tty:
        return ' '.join("{}{}{}".format(RED, c, CLEAR) for c in value)
    return ' '.join("{}".format(c) for c in value)


def cmd(value, tty=TTY):
    """Style the linux command string.

    :param value: the linux command string to style
    :param tty: True iff the output is going to a terminal
    :return: the styled command
    """
    return '{}{}{}'.format(GREEN, value, CLEAR) if tty else value


def bad(value, tty=TTY):
    """Style the invalid entry (e.g. directory path, tag, argument).

    :param value: the invalid value to style
    :param tty: True iff the output is going to a terminal
    :return: the styled value
    """
    return '{}{}{}'.format(RED, value, CLEAR) if tty else value


def msg(value, tty=TTY):
    """Style the message.

    :param value: the message to style
    :param tty: True iff the output is going to a terminal
    :return: the styled message
    """
    return '{}{}{}'.format(BLACK, value, CLEAR) if tty else value


def sign(value, tty=TTY):
    """Style the sign (e.g '+' and '-' symbols).

    :param value: the sign to style
    :param tty: True iff the output is going to a terminal
    :return: the styled sign
    """
    return '{}{}{}'.format(ORANGE_BOLD, value, CLEAR) if tty else value



CONFIGURATION = """
functions -e dtags > /dev/null 2>&1
functions -e t > /dev/null 2>&1
functions -e u > /dev/null 2>&1
functions -e e > /dev/null 2>&1
functions -e p > /dev/null 2>&1
functions -e d > /dev/null 2>&1

function dtags
    dtags-manage $argv; and dtags-refresh fish | source
end

function t
    dtags-t $argv; and dtags-refresh fish | source
end

function u
    dtags-u $argv; and dtags-refresh fish | source
end

function d
    set _dtags_usage "{usage}"
    set _dtags_version "{version}"
    set _dtags_description "{description}"
    if isatty 1
        set _dtags_arg_err "{arg_err_tty}"
        set _dtags_dest_err "{dest_err_tty}"
        set _dtags_input_err "{input_err_tty}"
        set _dtags_idx_err "{index_err_tty}"
        set _dtags_choice "{choice_tty}"
        set -g _dtags_prompt "{prompt_tty}"
    else
        set _dtags_arg_err "{arg_err}"
        set _dtags_dest_err "{dest_err}"
        set _dtags_input_err "{input_err}"
        set _dtags_idx_err "{index_err}"
        set _dtags_choice "{choice}"
        set -g _dtags_prompt "{prompt}"
    end
    if math "1 >" (count $argv) > /dev/null
        cd $HOME
        return 0
    else if [ $argv[1] = "--help" ]
        printf "$_dtags_usage$_dtags_description"
        return 0
    else if [ $argv[1] = "--version" ]
        printf "Version %s\n" "$_dtags_version"
        return 0
    else if [ $argv[1] = "-" ]
        cd "$OLDPWD"
        return 0
    else if echo "$argv[1]" | grep -q -E '^-' > /dev/null
        printf "$_dtags_arg_err" "$_dtags_usage" "$argv[1]"
        return 2
    else if math "1 <" (count $argv) > /dev/null
        printf "%sd: too many arguments\n" "$_dtags_usage"
        return 2
    end
    set _dtags_dirs (grep -F ,$argv[1], {mapping_file} | cut -d',' -f1)
    set -g _dtags_count (count $_dtags_dirs)
    if math "$_dtags_count == 0" > /dev/null
        if test -d $argv[1]
            set _dtags_dir $argv[1]
        else
            printf "$_dtags_dest_err" "$argv[1]"
            return 2
        end
    else if math "$_dtags_count == 1" > /dev/null
        set _dtags_dir $_dtags_dirs[1]
    else
        for i in (seq $_dtags_count)
            printf "$_dtags_choice" "$i" "$_dtags_dirs[$i]"
        end
        function _dtags_prompt_func
            printf "$_dtags_prompt" "$_dtags_count"
        end
        read -l -p _dtags_prompt_func _dtags_idx
        functions --erase _dtags_prompt_func
        if not echo "$_dtags_idx" | grep -q -E '^[0-9]+$' > /dev/null
            printf "$_dtags_input_err" "$_dtags_idx"
            return 2
        else if math "$_dtags_idx > $_dtags_count" > /dev/null
            printf "$_dtags_idx_err" "$_dtags_idx"
            return 2
        end
        set _dtags_dir $_dtags_dirs[$_dtags_idx]
    end
    if test -d $_dtags_dir
        cd $_dtags_dir
        return 0
    else
        printf "$_dtags_dest_err" "$_dtags_dir"
        return 2
    end
end

function __dtags_no_args
  set cmd (commandline -opc)
  if [ (count $cmd) -eq 1 ]
    return 0
  end
  return 1
end

function __dtags_one_arg
  set cmd (commandline -opc)
  if [ (count $cmd) -eq 2 ]
    return 0
  end
  return 1
end


function __dtags_has_first_arg
  set cmd (commandline -opc)
  if [ (count $cmd) -gt 1 ]
    if [ $argv[1] = $cmd[2] ]
      return 0
    end
  end
  return 1
end

function __dtags_past_first_arg
  set cmd (commandline -opc)
  if [ (count $cmd) -gt 1 ]
    return 0
  end
  return 1
end

function __dtags_tags
    if test -f {tags_file}
        cat {tags_file}
    end
end

complete -f -c dtags -n '__dtags_no_args' \
    -a list -d 'Display directories-to-tags mapping'
complete -f -c dtags -n '__dtags_no_args' \
    -a reverse -d 'Display tags-to-directories mapping'
complete -f -c dtags -n '__dtags_no_args' \
    -a edit -d 'Edit tags and directories via editor'
complete -f -c dtags -n '__dtags_no_args' \
    -a clean -d 'Remove invalid tags and directories'
complete -f -c dtags -n '__dtags_no_args' \
    -a shell -d 'Print the shell runtime configuration'
complete -f -c dtags -n '__dtags_no_args' \
    -a commands -d 'List all available dtags commands'
complete -f -c dtags -n '__dtags_has_first_arg list' \
    -a '(__dtags_tags)' -d 'Directory tag'
complete -f -c dtags -n '__dtags_has_first_arg reverse' \
    -a '(__dtags_tags)' -d 'Directory tag'

complete -f -c t -n '__dtags_past_first_arg' \
    -a '(__dtags_tags)' -d 'Directory tag'

complete -f -c u -n '__dtags_past_first_arg' \
    -a '(__dtags_tags)' -d 'Directory tag'

complete -f -c e -n '__dtags_no_args' \
    -a '(__dtags_tags)' -d 'Directory tag'

complete -f -c p -n '__dtags_no_args' \
    -a '(__dtags_tags)' -d 'Directory tag'

complete -f -c d -n '__dtags_no_args' \
-a '(__dtags_tags)' -d 'Directory tag'

dtags-refresh fish | source
"""



from __future__ import print_function

from setuptools import setup, find_packages
from setuptools.command.install import install

version = {}
with open('./dtags/version.py') as fp:
    exec(fp.read(), version)

with open('./README.rst') as fp:
    description = fp.read()


post_install_msg = """
Finished installing dtags {version}.
To activate dtags, place the following line in your shell config:

  For zsh, place in ~/.zshrc:
  command -v dtags-activate > /dev/null 2>&1 && eval "`dtags-activate zsh`"

  For bash, place in ~/.bashrc (or ~/.bash_profile for OS X):
  command -v dtags-activate > /dev/null 2>&1 && eval "`dtags-activate bash`"

  For fish, place in ~/.config/fish/config.fish:
  command -v dtags-activate > /dev/null 2>&1; and dtags-activate fish | source

And then restart your shell.""".format(version=version['VERSION'])


class DTagsInstall(install):
    def run(self):
        install.run(self)
        self.execute(lambda: None, [], msg=post_install_msg)

setup(
    name='dtags',
    version=version['VERSION'],
    description='Directory Tags for Lazy Programmers',
    long_description=description,
    author='Joohwan Oh',
    author_email='joohwan.oh@outlook.com',
    url='https://github.com/joowani/dtags',
    packages=find_packages(),
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Natural Language :: English',
        'Topic :: Utilities',
    ],
    cmdclass={'install': DTagsInstall},
    entry_points={
        'console_scripts': [
            'e = dtags.commands.e:main',
            'p = dtags.commands.p:main',
            'dtags-t = dtags.commands.t:main',
            'dtags-u = dtags.commands.u:main',
            'dtags-manage = dtags.commands:manage',
            'dtags-refresh = dtags.commands:refresh',
            'dtags-activate = dtags.commands:activate',
        ],
    }
)



from __future__ import unicode_literals

import os
import io
import errno
import shutil
from collections import defaultdict
from tempfile import NamedTemporaryFile

from dtags import style, CFG_DIR, TAGS_FILE, MAPPING_FILE
from dtags.chars import get_invalid_path_chars, get_invalid_tag_chars
from dtags.utils import abort, expand, rm_files


BULLET = '  > '


def parse_mapping(filename):
    """Parse the mapping file and return a mapping object.

    The mapping object looks like this (tag to directories):
    {
        'app' : {'/home/user/app/backend', '/home/app/frontend'},
        'frontend': {'/home/user/app/frontend'},
        'backend': {'/home/user/app/backend'},
    }
    Any invalid directories and tags are ignored.

    :param filename: the name of the mapping file to parse
    :return: the mapping object and set of ignored paths and tags
    """
    # Collect errors while parsing
    excluded = set()
    orphaned_paths = set()

    # The mapping object to populate
    mapping = defaultdict(set)

    # Parse the file line by line
    with io.open(filename, mode='rt') as mapping_file:
        for line in mapping_file.readlines():
            words = []
            for word in line.split(','):
                word = word.strip()
                if word:
                    words.append(word)
            # Skip empty lines or comments
            if not words or words[0].startswith('#'):
                continue
            # Collect orphan directories
            if len(words) == 1:
                orphaned_paths.add(words[0])
                continue
            path, tags = expand(words[0]), words[1:]
            valid_directory = True
            invalid_path_chars = get_invalid_path_chars(path)
            if invalid_path_chars:
                excluded.add(
                    BULLET + style.bad(path) +
                    ' contains bad characters: ' +
                    style.bad_chars(invalid_path_chars)
                )
                valid_directory = False
            elif not os.path.isdir(path):
                excluded.add(
                    BULLET + style.bad(path) +
                    ' is not a directory path'
                )
                valid_directory = False
            for tag in tags:
                if not tag[0].isalpha():
                    excluded.add(
                        BULLET + style.bad(tag) +
                        ' does not start with an alphabet '
                    )
                    continue
                if ' ' in tag:
                    excluded.add(
                        BULLET + style.bad(tag) +
                        ' contains whitespaces '
                    )
                    continue
                invalid_tag_chars = get_invalid_tag_chars(tag)
                if invalid_tag_chars:
                    excluded.add(
                        BULLET + style.bad(tag) +
                        ' contains bad characters: ' +
                        style.bad_chars(invalid_tag_chars)
                    )
                elif valid_directory:
                    mapping[tag].add(path)
                    orphaned_paths.discard(path)
        excluded.update(
            BULLET + style.bad(orphaned_path) +
            ' is not mapped to any valid tags'
            for orphaned_path in orphaned_paths
        )
    # Return the mapping and the ignored entries if any
    return mapping, '\n'.join(excluded) if excluded else None


def load_mapping():
    """Load the mapping from the disk.

    If the configuration directory files do not exist (probably because it is
    the user's first time executing a dtags command), create them.

    :return: the mapping object loaded from the disk
    """
    try:
        # Parse and load the mapping from the disk
        return parse_mapping(MAPPING_FILE)
    except (IOError, OSError) as mapping_load_error:
        if mapping_load_error.errno != errno.ENOENT:
            abort(
                message='Failed to read {file}: {msg}'.format(
                    file=MAPPING_FILE,
                    msg=mapping_load_error.strerror
                ),
                exit_code=mapping_load_error.errno
            )
        else:
            try:  # Create the directory if it does not exist
                os.makedirs(CFG_DIR)
            except (IOError, OSError) as dir_create_error:
                if dir_create_error.errno != errno.EEXIST:
                    abort(
                        message='Failed to create {file}: {msg}'.format(
                            file=CFG_DIR,
                            msg=dir_create_error.strerror
                        ),
                        exit_code=dir_create_error.errno
                    )
            for config_file_to_create in (MAPPING_FILE, TAGS_FILE):
                try:
                    open(config_file_to_create, 'w').close()
                except (IOError, OSError) as file_create_error:
                    if file_create_error.errno != errno.EEXIST:
                        abort(
                            message='Failed to create {file}: {msg}'.format(
                                file=config_file_to_create,
                                msg=file_create_error.strerror
                            ),
                            exit_code=file_create_error.errno
                        )
            return defaultdict(set), None  # Return an empty mapping


def save_mapping(mapping):
    """Save the mapping (and tags for autocomplete) back to the disk.

    This function assumes that the mapping is already validated.

    :param mapping: the mapping object to save
    """
    # Create a reverse mapping (this is what gets saved to the disk)
    reverse_mapping = defaultdict(set)
    for tag, paths in mapping.items():
        for path in paths:
            reverse_mapping[path].add(tag)
    # Save the reverse mapping to a temporary file
    temp_mapping_file = NamedTemporaryFile(
        mode='w', dir=CFG_DIR, prefix='mapping.', delete=False
    )
    temp_mapping_file.close()
    try:
        with io.open(temp_mapping_file.name, 'wt') as temp_mapping_file:
            temp_mapping_file.write('\n'.join(
                path + ',' + ','.join(sorted(reverse_mapping[path])) + ','
                for path in sorted(reverse_mapping)
            ))
            temp_mapping_file.flush()
            os.fsync(temp_mapping_file.fileno())
    except (IOError, OSError) as temp_mapping_write_error:
        rm_files(temp_mapping_file.name)
        abort(
            message='Failed to write to {name}: {msg}\n'.format(
                name=temp_mapping_file.name,
                msg=temp_mapping_write_error.strerror
            ),
            exit_code=temp_mapping_write_error.errno
        )
    # Save the tags (for tab-completion) to a temporary file
    temp_tags_file = NamedTemporaryFile(
        mode='w', prefix='tags.', dir=CFG_DIR, delete=False
    )
    temp_tags_file.close()
    try:
        with io.open(temp_tags_file.name, 'wt') as temp_tags_file:
            temp_tags_file.write('\n'.join(sorted(mapping)))
            temp_tags_file.flush()
            os.fsync(temp_tags_file.fileno())
    except (IOError, OSError) as temp_tags_write_error:
        rm_files(temp_mapping_file.name, temp_tags_file.name)
        abort(
            message='Failed to write to {name}: {msg}\n'.format(
                name=temp_tags_file.name,
                msg=temp_tags_write_error.strerror
            ),
            exit_code=temp_tags_write_error.errno
        )
    try:  # Overwrite the mapping file with the temporary one
        shutil.move(temp_mapping_file.name, MAPPING_FILE)
    except (IOError, OSError) as mapping_rename_error:
        rm_files(temp_mapping_file.name, temp_tags_file.name)
        abort(
            message='Failed to move {src} to {dest}: {msg}\n'.format(
                src=temp_mapping_file.name,
                dest=MAPPING_FILE,
                msg=mapping_rename_error.strerror
            ),
            exit_code=mapping_rename_error.errno
        )
    try:  # Overwrite the tags file with the temporary one
        shutil.move(temp_tags_file.name, TAGS_FILE)
    except (IOError, OSError) as tags_rename_error:
        rm_files(temp_tags_file.name)
        abort(
            message='Failed to move {src} to {dest}: {msg}\n'.format(
                src=temp_tags_file.name,
                dest=TAGS_FILE,
                msg=tags_rename_error.strerror
            ),
            exit_code=tags_rename_error.errno
        )
